//
//                     Pervasive Discovery Protocol
//                          Copyright (C) 2005
//
//    Jos Carlos Perea barreiro         Celeste Campo Vzquez
//      (jperea[at]it.uc3m.es)          (celeste[at]it.uc3m.es)
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
//
// Contact the authors at:
//
//    Jos Carlos Perea Barreiro (jperea [at]it.uc3m.es)
//    Celeste Campo Vazquez      (celeste[at]it.uc3m.es)
//
//        Universidad Carlos III de Madrid
//        Departamento de Ingeniera Telemtica
//        Av. Universidad, 30
//        Edif. Torres Quevedo.
//        (28911) Legans - Madrid - SPAIN
//
//
/////////////////////////////////////////////////////////////////////////////

/**
 * Provides the classes and interfaces relating to the
 * reference implementation for Pervasive Discovery Protocol
 * (PDP).
 * <p>
 * This protocol was designed within the Departamento de
 * Ingeniera Telemtica at the Universidad Carlos III de
 * Madrid by Celeste Campo Vzquez (celeste[at]it.uc3m.es)
 * and programmed by Jos Carlos Perea Barreiro
 * (jperea[at]it.uc3m.es)
 */

package com.uc3m.it.pfc.pdp;

import java.util.Vector;

/**
 * The PDP class is the core class for this PDP protocol
 * implementation. It provides access to the discovery
 * protocol from a device, so that it can control User and
 * Service Agents, and manage and access both local and
 * remote cache memory. <br>
 * <br>
 *
 * Ver 0.1: Multithreading improved
 *
 * @author Celeste Campo Vzquez (celeste[at]it.uc3m.es)
 * @author Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
 * @version 0.1
 */

public class PDP
{
    /**
     * Sets maximun simultaneus in-connections permitted
     */
    private final static byte MAX_CONNECTIONS = 10;

    /**
     * Sets maximun time (in seconds) UA would wait for a
     * complete reply
     */
    protected final static int CONFIG_WAIT_REPLY = 15;

    /**
     * Sets maximun time (in milliseconds) a Client would
     * wait for a complete reply. By default it is set as
     * CONFIG_WAIT_REPLY + 5. Clients would use this
     * parameter to wait before accessing a response from a
     * SD proccess.
     */
    public final static int CLIENT_WAIT = ( CONFIG_WAIT_REPLY + 5 ) * 1000;

    /**
     * Set maximun waitting time (in seconds) for a device
     * to reply a ServiceRequest message. It MUST be smaller
     * than CONFIG_WAIT_REPLY
     */
    public final static int CONFIG_WAIT_TRANSMIT = 10;

    /**
     * Set maximun time (in seconds) since a device is seen
     * as static
     */
    public final static int DEFAULT_AVAILABILITY_TIME = 3600;

    /**
     * User agent. Thread responsible for discovering
     * services
     *
     * @uml.property name="ua"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private PDP_UA ua;

    /**
     * Service agents. Group of threads responsible for
     * announcing services
     *
     * @uml.property name="sa"
     * @uml.associationEnd multiplicity="(0 -1)"
     */
    private PDP_SA [] sa;

    /**
     * Network Agent. Thead responsible for sending and
     * receiving messages
     *
     * @uml.property name="nm"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private PDPNetworkManager nm;

    /**
     * Local cache. Place to store services held by the
     * device.
     *
     * @uml.property name="local"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private PDPCache local;

    /**
     * Remote cache. Place to store services learnt from
     * network.
     *
     * @uml.property name="cache"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private PDPCache cache;

    /**
     * Flag to allow UserAgent activity. When -
     * <code>true</code>-, the device can search and
     * discover services.
     */
    private boolean ua_active;

    /**
     * Flag to allow ServiceAgent activity. When -
     * <code>true</code>-, the device can offer his
     * services and interact with other devices.
     */
    private boolean sa_active;

    /**
     * Flag to force device to announce when he is leaving
     * the network. When -<code>true</code>-, device
     * sends a ServiceDeregister message whenever is leaving
     * a network where it offers services
     */
    private boolean deregister;

    /**
     * Flag active whenever there is a new completed reply
     * for a request sent form this device. Modified on
     * 2005-03-28
     */
    private boolean new_data;

    /**
     * Flag to control UA work. When active UA is working on
     * a service discovery and wont attend any other service
     * request. Added on 2005-03-28
     */
    private boolean ua_busy;

    /**
     * Holds the device availability time (in seconds), the
     * time that this device estimates it is being active on
     * the network
     */
    private int pdp_time;

    /**
     * Holds the XID message field. Identifies every request
     * message sent through the network
     */
    private int xid;

    /**
     * Stores ServiceEntry objects received as reply for the
     * last request sent from this device
     */
    private Vector reply;

    /**
     * Object where notify the result for a new SD proccess.
     * See {@link searchService ( String, boolean  )}
     */
    private Object SDnotifier;

    /**
     * Object where notify changes on the network.
     * See {@link getNetworkLock}
     */
    private Object NTnotifier;


    /**
     * Lock for synchronizing UA data
     */
    private Object UAlock;

    /**
     * Lock for synchronizing SA data
     */
    private Object SAlock;

    /**
     * Lock for synchronizing {@link PDP#deregister}
     */
    private Object DElock;

    /**
      * IP address for this device. <code>null</code> if
      * there is no network connection available
      */
    private String localhostadress;


    /**
     *
     * Constructor. Configures the device so that it can
     * start offering/discovering services on the network.
     *
     *
     * @param local_length
     *            Maximun number of elements (
     *            {@link PDPCacheData}) held on local cache
     *            memory
     *
     * @param cache_length
     *            Maximun number of element (
     *            {@link PDPCacheData}) held on remote
     *            cache memory
     *
     * @param pdp_time
     *            availability time (in seconds) for this
     *            device
     *
     * @throws PDPException
     *             If an network error happens
     *
     */
    public PDP ( int local_length, int cache_length, int pdp_time ) throws PDPException
    {
        if ( pdp_time > 65535 ) pdp_time = 65535;

        this.sa = null;
        this.ua = null;

        this.xid = ( int ) ( Math.random ( ) * 65534 );

        this.pdp_time = pdp_time;

        this.local = new PDPCache ( this, local_length, true, pdp_time );
        this.cache = new PDPCache ( this, cache_length, false, pdp_time );

        this.ua_active  = false;
        this.sa_active  = false;
        this.deregister = false;
        this.new_data   = false;
        this.ua_busy    = false;

        this.UAlock     = new Object ( );
        this.SAlock     = new Object ( );
        this.DElock     = new Object ( );
        this.SDnotifier = new Object ( );
        this.NTnotifier = new Object ( );

        this.reply = null;
        this.nm    = new PDPNetworkManager ( this, NTnotifier );
    }

    /**
     * Enables device to discover services. When flag
     * <code>active</code> is <code>true</code> device
     * can use his User Agent {@link PDP_UA}to discover
     * services on the network.
     *
     * @param active
     *            Flag to enable device to discover services
     */
    public void setUA ( boolean active )
    {
        synchronized ( UAlock )
        {
            try
            {
                if ( !active ) killUA ( );
            }
            catch ( Exception e )
            { /* doesn't mind since we are killing UA */
            }

            ua_active = active;
        }
    }

    /**
     * Enables device to offer services. When flag
     * <code>active</code> is <code>true</code> device
     * can use his Servuce Agent {@link PDP_SA}to offer
     * services to the network.
     *
     * @param active
     *            Flag to enable device to offer services
     */
    public void setSA ( boolean active )
    {
        synchronized ( SAlock )
        {
            if ( active )
            {
                sa = new PDP_SA [MAX_CONNECTIONS];
            }
            else
            {
                sa = null;
            }

            sa_active = active;
        }
    }

    /**
     * Configures device to inform other devices whenever it
     * is leaving the network by using a ServiceDeregister
     * message
     *
     * @param active
     *            Flag to configure device to sending
     *            ServiceDeregister messages when he is
     *            leaving the network
     *
     */
    public void setDereg ( boolean active )
    {
        synchronized ( DElock )
        {
            deregister = active;
        }
    }

    /**
     * Adds an entry on local cache with the information
     * given on the parameters.
     *
     * @param service
     *            {@link PDPService}to add on local cache
     * @param ttl
     *            expiration time for this service on local
     *            cache
     *
     * @return <code>true</code> when the entry has been
     *         correctly added on the cache
     *
     */
    public boolean addLocal ( String service, String url, int ttl )
            throws PDPException
    {
        ServiceURL u = new ServiceURL ( service, url );
        return local.add ( u, ttl );
    }

    /**
     * Launches a service discovery request for the service
     * <code>type</code>.
     *
     * @param type
     *            Type of service to search
     * @param any
     *            <code>true</code> means 1-1 request type, <code>false</code>
     *            means 1-N request type. See PDP documentation for
     *            further information
     *
     * @return synchronizing monitor where PDP will notify there
     *         is an available reply
     */
    public Object searchService ( String type, boolean any ) throws PDPException
    {
        synchronized ( UAlock )
        {
            if ( ua_active )
            {
                if ( !ua_busy )
                {
                    new_data = false;
                    ua_busy = true;
                    reply = null;
                    xid++;
                    ua = new PDP_UA ( this, type, xid, any );
                }
                else
                {
                    throw new PDPException ( PDPException.UA_BUSY );
                }
            }
            else
            {
                throw new PDPException ( PDPException.UA_DISABLED );
            }
        }

        return SDnotifier;
    }

    /**
     * Gets (if available) the response for the last SD
     * process. When invoked it always stops the UA.
     *
     * @return Vector of ServiceEntries data as a response
     *         for a SD process
     * @throws PDPException
     *             Type UA_DISABLED if User Agent is
     *             disabled
     */
    public Vector getResponse ( ) throws PDPException
    {
        Vector data = null;

        synchronized ( UAlock )
        {
            if ( !ua_active )
            {
                throw new PDPException ( PDPException.UA_DISABLED );
            }

            if ( new_data )
            {
                data = reply;
                new_data = false;

                killUA ( );
            }
        }

        return data;
    }

    /*
     * Reads the information on <code> Vector reply </code>
     * and returns an array of {@link java.lang.String}
     * objects with the response for a request made by this
     * device. If there is no response, it returns a
     * one-string array with the String <code> "NOT FOUND"
     * </code> . Otherwise it returns the info by using the
     * following template: <p>
     *
     * <pre> expiration_time + &quot;secs &quot; +
     * service_description </pre>
     *
     * @return Array of {@link java.lang.String}objects.
     */
    /*
     * // INITIAL VERSION ---> REMOVED [2005-03-28] public
     * String [] getResponse ( ) { String [] st = null;
     * synchronized ( reply ) { if ( reply.size ( ) == 0 ) {
     * st = new String [1]; st[0] = "NOT FOUND"; } else {
     * int j = reply.size ( ); st = new String [j]; for (
     * int i = 0; i < reply.size ( ); i ++ ) { ServiceEntry
     * s = (ServiceEntry)reply.elementAt (i); String desc =
     * s.getPDPService().getDescription(); int ttl =
     * s.getLifetime ( ); st [i] = ttl + "secs " + desc; } }
     * reply = null; }
     *
     * return st; }
     */

    /**
     * Returns the content of both local and remote cache.
     *
     * @return Array of {@link ServiceEntry}objects with
     *         both local and remote cache content.
     */
    /*
     * // INITIAL VERSION --> REMOVED [2005-03-28] public
     * String [] showCache ( ) { Vector v = new Vector();
     * v.addAll ( local.getServices ( "all" ) ); v.addAll (
     * cache.getServices ( "all" ) ); reply ( v ); while (
     * !new_data ); return getResponse ( ); }
     */
    // Version 0.1 [2005-03-28]
    public Vector showCache ( )
    {
        Vector v = new Vector ( );
        v.addAll ( local.getServices ( "all" ) );
        v.addAll ( cache.getServices ( "all" ) );
        return v;
    }

    /**
     * Stops the current job for the UA
     *
     * @throws PDPException
     *             PDPException.UA_DISABLED if UA is not
     *             enabled
     */
    public void stopUA ( ) throws PDPException
    {
        synchronized ( UAlock )
        {
            if ( !ua_active )
            {
                throw new PDPException ( PDPException.UA_DISABLED );
            }

            if ( ua_busy )
            {
                ua.stopUA ( );
                new_data = false;
                killUA ( );

                synchronized ( SDnotifier )
                {
                    SDnotifier.notify ( );
                }
            }
        }
    }

    /**
      * Returns the synchronizing monitor where PDP will notify
      * about network connectivity
      */
    public Object getNetworkLock ( )
    {
        return NTnotifier;
    }

    /**
      * Gets the IP address of this device or <code>null</code> if it's
      * not connected to an active netowrk
      *
      * @return IP address of this device.
      */    public String getLocalhostAddress ( )
    {
        return localhostadress == null ? "0.0.0.0" : localhostadress;
    }

    /**
      * Informs about netowork connectivity.
      *
      * @return <code>true</code> if PDP is connected to an active network
      */
    public boolean isConnected( )
    {
        return ( localhostadress != null );
    }

    /**
      * Stops PDP proccess and closes/frees all allocated resoruces
      */
    public void byebye ( )
    {
        deregister ( );
        try
        {
            stopUA ( );
        }
        catch ( Exception e )
        {
        }
        
        nm.stopLoop ( );
        nm.stopSniffer ( );
        
        synchronized ( NTnotifier )
        {
        	NTnotifier.notify();        	
        }
        
        ua_active  = false;
        sa_active  = false;
        deregister = false;
        SDnotifier = null;
        ua         = null;
        sa         = null;
        nm         = null;
        UAlock     = null;
        SAlock     = null;
        DElock     = null;
        reply      = null;
        local      = null;
        pdp_time   = 0;
        xid        = 0;

        System.gc ( );
    }

    /**
      *
      */
    protected void networkInformation ( String localhostaddress )
    {
        this.localhostadress = localhostaddress;
    }

    /**
     * Returns the availability time of this device (in
     * seconds)
     *
     * @return availability time (in seconds) for this
     *         device
     */
    protected int getTTL ( )
    {
        return pdp_time;
    }

    /**
     * Returns a reference to either local/remote cache of
     * this device. This method is used by {@link PDPCache}
     * before adding a service to access to other the other
     * cache memory and check whether the service is already
     * stored.
     *
     * @param local
     *            Indicates whether you need local -
     *            <code>true</code>- or remote -
     *            <code>false</code>- cache.
     *
     * @return {@link PDPCache}object with a reference for
     *         the local/remote cache memory of the device
     */
    protected PDPCache getCache ( boolean local )
    {
        if ( local )
            return this.local;
        else
            return this.cache;
    }

    /**
     * Removes the services stored on the array of
     * {@link ServiceEntry}objects from the remote cache.
     * This method is used by {@link PDPNetworkManager}when
     * it receives a ServiceDeregister message
     *
     * @param data
     *            Array of {@link ServiceEntry}objects
     *            which contains the services to be removed
     *            from remote memory
     */
    protected synchronized void remove ( ServiceEntry [] data )
    {
        cache.remove ( data );
    }

    /**
     * Adds the services stored on the array of
     * {@link ServiceEntry}objects to remote cache. This
     * method is used by {@link PDPNetworkManager}when it
     * receives either a ServiceRequest/ServiceReply
     * message.
     *
     * @param data
     *            Array of {@link ServiceEntry}objects
     *            which contains the services to be added to
     *            remote memory
     */
    protected synchronized void add ( ServiceEntry [] data )
    {
        cache.add ( data );
    }

    /**
     * Notifies PDP core that a new reply message has been
     * received. This method is used by
     * {@link PDPNetworkManager}when it receives a new
     * ServiceReply message so it can use the information on
     * the message to complete a request sent from the
     * device, or modify a reply that is being prepared
     *
     * @param msg
     *            {@link PDPMessage}received from network
     */
    protected synchronized void newReply ( PDPMessage msg )
    {
        int _xid = msg.getXID ( );

        synchronized ( UAlock )
        {
            if ( ua_active )
            {
                if ( _xid == xid )
                {
                    if ( ua_busy )
                    {
                        ua.newData ( msg.getServiceEntry ( ) );
                    }
                }
            }
        }
        synchronized ( SAlock )
        {
            if ( sa_active )
            {
                if ( _xid != xid )
                {
                    for ( int i = 0; i < MAX_CONNECTIONS; i++ )
                    {
                        if ( sa[i] != null )
                        {
                            int __xid = Integer.parseInt ( sa[i].getName ( ) );

                            if ( __xid == _xid )
                            {
                                sa[i].newData ( msg.getServiceEntry ( ) );
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Notifies PDP core that a new request message has been
     * received. This method is used by
     * {@link PDPNetworkManager}when it receives a new
     * ServiceRequest message, so it can start a new service
     * offering proccess
     *
     * @param msg
     *            {@link PDPMessage}received from network
     */
    protected void newRequest ( PDPMessage msg )
    {
        if ( msg.getXID ( ) == xid )
        {
            return;
        }

        synchronized ( SAlock )
        {
            if ( sa_active )
            {
                for ( int i = 0; i < MAX_CONNECTIONS; i++ )
                {
                    if ( sa[i] == null
                            || ( sa[i] != null && !sa[i].isAlive ( ) ) )
                    {
                        sa[i] = new PDP_SA ( this, msg );
                        i = MAX_CONNECTIONS;
                    }
                }
            }
        }
    }

    /**
     * Sends a {@link PDPMessage}to
     * {@link PDPNetworkManager}so that it can send it
     * through the network. This method is used by
     * {@link PDP_UA}and {@link PDP_SA}objects when they
     * need to send a message to other devices.
     *
     * @param msg
     *            Message to be sent throght the network
     */
    protected synchronized void toNetworkManager ( PDPMessage msg )
    {
        // Sends new data to NetworkManager
        nm.newData ( msg );
    }

    /**
     * Receives a response from a service discovery process.
     * This method is used by {@link PDP_UA}whenever it has
     * finish its discovering process so that PDP core gets
     * informed that new data is available.
     *
     * @param data
     *            Vector of {@link ServiceEntry}objects
     *            received from a service discovering
     *            process
     */
    protected void reply ( Vector data )
    {
        synchronized ( UAlock )
        {
            reply = data;
            new_data = true;

            synchronized ( SDnotifier )
            {
                SDnotifier.notify ( );
            }
        }
    }

    /**
     * Resets {@link PDP_SA}memory when a service offering
     * process has finished. This method is used by the
     * {@link PDP_SA}object who finished its work so that
     * PDP core can free this memory to use it on the next
     * reply
     */
    protected void killSA ( String xid )
    {
        synchronized ( SAlock )
        {
            if ( sa_active )
            {
                for ( int i = 0; i < MAX_CONNECTIONS; i++ )
                {
                    if ( sa[i] != null && sa[i].getName ( ).equals ( xid ) )
                    {
                        sa[i] = null;
                    }
                }
            }
        }
    }

    /**
     * Resets {@link PDP_UA}memory when either the device
     * has finished or has been stopped from an SD proccess.
     */
    private void killUA ( ) throws PDPException
    {
        if ( ua_active )
        {
            ua = null;
            ua_busy = false;
        }
        else
        {
            throw new PDPException ( PDPException.UA_DISABLED );
        }
    }

    /**
     * Creates (if enabled) a ServiceDeregister message and
     * sends it to {@link PDPNetworkManager}to send it to
     * every device available on the network.
     */
    private void deregister ( )
    {
        synchronized ( DElock )
        {
            if ( deregister )
            {
                Vector v = local.getServices ( "all" );
                try
                {
                    PDPMessage msg = new PDPMessage ( v );

                    toNetworkManager ( msg );
                    
                    DElock.wait ( 1000 );
                }
                catch ( PDPException e )
                {
                }
                catch ( Exception e )
                {
                }
            }
        }
    }
}