//
//                     Pervasive Discovery Protocol
//                          Copyright (C) 2005
//
//    Jos Carlos Perea barreiro         Celeste Campo Vzquez
//      (jperea[at]it.uc3m.es)          (celeste[at]it.uc3m.es)
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
//
// Contact the authors at:
//
//    Jos Carlos Perea Barreiro (jperea [at]it.uc3m.es)
//    Celeste Campo Vazquez      (celeste[at]it.uc3m.es)
//
//        Universidad Carlos III de Madrid
//        Departamento de Ingeniera Telemtica
//        Av. Universidad, 30
//        Edif. Torres Quevedo.
//        (28911) Legans - Madrid - SPAIN
//
//
/////////////////////////////////////////////////////////////////////////////

/**
  * Provides the classes and interfaces relating to the reference
  * implementation for Pervasive Discovery Protocol (PDP).
  * <p>
  * This protocol was designed within the Departamento de Ingeniera
  * Telemtica at the Universidad Carlos III de Madrid by
  * Celeste Campo Vzquez (celeste[at]it.uc3m.es) and programmed by
  * Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
  */
package com.uc3m.it.pfc.pdp;

import java.util.Vector;
import java.util.Collections;

/**
 * The PDPCache class implements a cache memory stored on a PDP device.
 * <p>
 * This memory can be a <i>local cache memory</i>, to store services
 * held by the device, or <i>remote cache memory</i>, to store services
 * learnt from the network. The basic difference beween these memories
 * is found in its elemets; a service stored on a remote cache can
 * expire, so this class implements a mechanism to auto-update
 * remote cache memory and remove its invalid service entries.
 *
 *
 *
 * @author Celeste Campo Vzquez (celeste[at]it.uc3m.es)
 * @author Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
 * @version initial
 */

class PDPCache extends Vector
{

    /**
     * Reference to PDP core, to access other agents and cache memory of
     * the device
     *
     * @uml.property name="padre"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private PDP padre;

    /**
      * Flag to specify whether it is a local -<code>true</code>- or
      * remote -<code>false</code>- cache memory
      */
    private boolean local = false;

    /**
      * Maximun number of entries to store on this memory
      */
    private int max_size;

    /**
      * Availability time of this device
      */
    private int pdp_time;



    /**
      * Constructor. Creates and configures a new PDP cache memory.
      *
      * @param padre Reference to PDP core to access other agents and
      *     cache memory of this device
      * @param max_size Maximun number of PDPServices permitted
      *     within this cache memory.
      * @param local boolean to set whether this memory is acting as
      *     local -<code>true</code>- or remote -<code>false</code>
      *     memory
      * @param pdp_time Availability time (in seconds) of this device
      */
    protected PDPCache ( PDP padre, int max_size, boolean local, int pdp_time )
    {
        this.padre = padre;
        this.local = local;
        this.max_size = max_size;
        this.pdp_time = pdp_time;
    }



    /**
      * Adds all services given on the array parameter. It internaly uses
      * the method {@link PDPCache#add(PDPService, int)}.
      *
      * @param entry Array of ServiceEntries to add in this memory
      */
    protected void add ( ServiceEntry [] entry )
    {
        if ( max_size == 0 ) return;

        int j = entry.length;

        for ( int i = 0; i < j; i++ )
        {
            if ( entry[i] != null && !entry [i].isNoCaching ( ) )
            {
                add ( entry[i].getPDPService ( ), entry[i].getLifetime () );
            }
        }
    }




    /**
      * Adds a new {@link PDPService} in the cache memory if it is not already
      * stored on this cache or in the PDP local cache. This method is called
      * by {@link PDPNetworkManager} whenever it receives a request/reply so
      * that device can learn form network
      *
      * @param service {@link PDPService} to add in this memory
      * @param ttl Expiration time for this {@link PDPService}
      * @return boolean <code>true</code> if the service has been added,
      *         <code>false</code> otherwise
      */
    protected boolean add ( PDPService service, int ttl )
    {
        if ( max_size == 0 ) return false;

        // If I'm remote cache, and local one has the PDPService "service"
        if ( !local && padre.getCache ( true ).hasPDPService ( service ) != -1 )
        {
            // DO NOT add this entry
            return false;
        }

        boolean r = false;

        if ( hasPDPService ( service ) == -1 )
        {
            if ( ttl > pdp_time ) ttl = pdp_time;

            PDPCacheData p = new PDPCacheData ( service, ttl, local );
            add ( p );
            Collections.sort ( this );
            if ( size ( ) > max_size ) removeElementAt ( 0 );

            r = true;
        }

        return r;
    }



    /**
      * Gets the {@link PDPCacheData} stored on the specified position in
      * this memory
      *
      * @param i Returns the PDPCacheData at the specified index
      * @return PDPCacheData element from the cache or <code>null</code>
      *     if it is an invalid element
      */
    private PDPCacheData getData ( int i ) throws ClassCastException
    {
        PDPCacheData data;

        if ( i < 0 || i > size ( ) ) data = null;
        else                         data = (PDPCacheData) elementAt( i );

        return data;
    }



    /**
      * Checks memory and retrieves all Services that match with the
      * service type specified on the parameter.
      *
      * @param service_type Type of service to retrieve from de cache memory
      * @return Vector Vector of {@link ServiceEntry} with the specified
      *     type of services
      */
    protected Vector getServices ( String service_type )
    {
        if ( max_size == 0 ) return new Vector ( );

        refresh ( );

        Vector v = new Vector ( );

        synchronized ( this )
        {
            int i = size();

            for ( int j = 0; j < i; j++ )
            {
                try
                {
                    PDPCacheData data = getData ( j );

                    if ( ( data.hasType ( service_type ) ) )
                    {
                        v.add ( data.toServiceEntry ( ) );
                    }
                }
                catch ( ClassCastException e )
                {
                    remove ( j );
                    j--;
                    i--;
                }
            }
        }

        v.trimToSize ( );

        return v;
    }


    /**
      * Checks is the {@link PDPService} <code>service</code> is already
      * stored on this memory
      *
      * @param service PDPService to check whether it is saved in the cache
      * @return Position of the service on this memory, or <code>-1</code>
      *     if the service is not found.
      */
    private int hasPDPService ( PDPService service )
    {
        refresh ( );

        int w = -1;

        synchronized ( this )
        {
            int i = size ( );

            for ( int j = 0; j < i; j++ )
            {
                if ( getData ( j ).hasService ( service ) )
                {
                    w = j;
                    j = i;
                }
            }
        }

        return w;
    }


    /**
      * Removes the services specified on the parameter form this memory.
      * This method is called by {@link PDPNetworkManager} whenever it
      * receives a ServiceDeregister message.
      *
      * @param entry Array of {@link ServiceEntry} objects with the
      *     services to be removed from this memory
      */
    protected void remove ( ServiceEntry [] entry )
    {
        if ( max_size == 0 ) return;

        int j = entry.length;

        for ( int i = 0; i < j; i++ )
        {
            remove ( entry[i].getPDPService ( ) );
        }
    }


    /**
      * Removes the specified service (if found) from this memory.
      *
      * @param service {@link PDPService} to remove from this memory.
      */
    private void remove ( PDPService service )
    {
        if ( service.getService ( ).equals ( "all" ) ) return;

        synchronized ( this )
        {
            int i = hasPDPService ( service );

            if ( i != -1 ) remove ( i );
        }
    }


    /**
      * Updates memory and removes expired data if and only if this object
      * represents a remote cache. Elements stored on a local cache do not
      * expire, they are services held on the device.
      */
    private void refresh ( )
    {
        if ( local ) return;

        synchronized ( this )
        {
            int i = size ( );
            int j = 0;
            boolean fin = false;

            // Searches for the first not expired element
            for ( j = 0; j < i && !fin; j ++ )
            {
                if ( !getData ( j ).hasExpired ( ) )
                {
                    fin = true;
                    j --;
                }
            }

            // Removes all expired elements
            if ( j > 0 ) { removeRange ( 0, j ); }
        }
    }
}
