//
//                     Pervasive Discovery Protocol
//                          Copyright (C) 2005
//
//    Jos Carlos Perea barreiro         Celeste Campo Vzquez
//      (jperea[at]it.uc3m.es)          (celeste[at]it.uc3m.es)
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
//
// Contact the authors at:
//
//    Jos Carlos Perea Barreiro (jperea [at]it.uc3m.es)
//    Celeste Campo Vazquez      (celeste[at]it.uc3m.es)
//
//        Universidad Carlos III de Madrid
//        Departamento de Ingeniera Telemtica
//        Av. Universidad, 30
//        Edif. Torres Quevedo.
//        (28911) Legans - Madrid - SPAIN
//
//
/////////////////////////////////////////////////////////////////////////////

/**
  * Provides the classes and interfaces relating to the reference
  * implementation for Pervasive Discovery Protocol (PDP).
  * <p>
  * This protocol was designed within the Departamento de Ingeniera
  * Telemtica at the Universidad Carlos III de Madrid by
  * Celeste Campo Vzquez (celeste[at]it.uc3m.es) and programmed by
  * Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
  */
package com.uc3m.it.pfc.pdp;

/**
 * The PDPCacheData class implements the objects stored on a PDP
 * cache memory.
 * <p>
 * Each element has a {@link PDPService} which represents a service
 * description, and a <code>expiration_time</code> which as you
 * can image, contains the information about the time this service
 * is valid and active.
 *
 *
 * @author Celeste Campo Vzquez (celeste[at]it.uc3m.es)
 * @author Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
 * @version initial
 */

class PDPCacheData implements Comparable
{

    /**
     * Service information stored in the memory
     *
     * @uml.property name="service"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private PDPService service;

    /**
      * Expiration time for this service
      */
    private long expiration_time;

    /**
      * Flag to specify whether it is a local -<code>true</code>- or
      * remote -<code>false</code>- cache memory entry
      */
    private boolean local;




    /**
      * Constructor. Creates a new entry for a PDP cache memory with the
      * information given on the parameters.
      *
      * @param service {@link PDPService} to store on the memory
      * @param ttl expiration time for this service
      * @param local boolean specifying if this service is stored as a
      *     local (static) -<code>true</code>- entry or a remote
      *     -<code>false</code>- entry.
      */
    protected PDPCacheData ( PDPService service, long ttl, boolean local )
    {
        this.service = service;
        this.local   = local;
        ttl = ttl * 1000;

        if ( !local )
            this.expiration_time = System.currentTimeMillis() + ttl;

        else
            this.expiration_time = ttl;
    }




    /**
      * Checks if this entry has expired. When this entry is stored as a
      * local cache memory data it never expires, so it returns
      * <code>false</code>.
      *
      * @return a boolean specifying whether this entry has expired or not
      */
    protected boolean hasExpired ( )
    {
        if ( local ) return false;
        else         return ( System.currentTimeMillis ( ) >= expiration_time );
    }



    /**
      * Checks if the service stored on this entry matches the service type
      * given on the parameter. It internaly uses method
      * {@link com.uc3m.it.pfc.pdp.PDPService#equals(String)}
      *
      * @param service_type Type of service to check with this entry
      * @return a boolean specifying whether this service matches with
      *     the type of service given on the parameter
      */
    protected boolean hasType ( String service_type )
    {
        boolean x = this.service.equals ( service_type );
        boolean y = !hasExpired ( );
        boolean z = x && y;
        return z;
    }



    /**
      * Checks if the service stored on this entry matches the service
      * given on the parameter. A service matches with another when
      * their service description is byte-to-byte equal. This method
      * internaly uses method
      * {@link com.uc3m.it.pfc.pdp.PDPService#equals(PDPService)}
      *
      * @param service {@link PDPService} to compare with this entry
      * @return a boolean specifying whether this service matches with
      *     the service given on the parameter
      */
    protected boolean hasService ( PDPService service )
    {
        return !hasExpired ( ) && this.service.equals ( service );
    }




    /**
      * Translates the information given on this entry into a
      * {@link ServiceEntry} object.
      *
      * @return {@link ServiceEntry} object with the information
      *     from this cache memory entry
      */
    protected ServiceEntry toServiceEntry ( )
    {
        long ttl;

        if ( local )
            ttl = this.expiration_time / 1000;

        else
            ttl = ( this.expiration_time - System.currentTimeMillis () ) / 1000;

        return new ServiceEntry ( service, ttl );
    }



    /**
      * Compares two PDPCacheData objects to be sorted from minor to mayor
      * expiration time.
      *
      * @see java.lang.Comparable#compareTo(Object)
      */
    public int compareTo(Object data)
    {
        PDPCacheData c = (PDPCacheData)data;

        return (int)(expiration_time - c.expiration_time);
    }
}