//
//                     Pervasive Discovery Protocol
//                          Copyright (C) 2005
//
//    Jos Carlos Perea barreiro         Celeste Campo Vzquez
//      (jperea[at]it.uc3m.es)          (celeste[at]it.uc3m.es)
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
//
// Contact the authors at:
//
//    Jos Carlos Perea Barreiro (jperea [at]it.uc3m.es)
//    Celeste Campo Vazquez      (celeste[at]it.uc3m.es)
//
//        Universidad Carlos III de Madrid
//        Departamento de Ingeniera Telemtica
//        Av. Universidad, 30
//        Edif. Torres Quevedo.
//        (28911) Legans - Madrid - SPAIN
//
//
/////////////////////////////////////////////////////////////////////////////

/**
  * Provides the classes and interfaces relating to the reference
  * implementation for Pervasive Discovery Protocol (PDP).
  * <p>
  * This protocol was designed within the Departamento de Ingeniera
  * Telemtica at the Universidad Carlos III de Madrid by
  * Celeste Campo Vzquez (celeste[at]it.uc3m.es) and programmed by
  * Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
  */
package com.uc3m.it.pfc.pdp;

import java.util.Vector;

/**
 * The PDPMessage class implements the mechanism to create, manage and parse
 * all available PDP messages.
 * <p>
 * It provides mechanisms to create new messages from basic data types,
 * convert it into byte arrays to adapt them to be sent through the network,
 * and read an parse messages received from the network.
 *
 *
 * @author Celeste Campo Vzquez (celeste[at]it.uc3m.es)
 * @author Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
 * @version initial
 */

class PDPMessage
{
    //////////////////////////////////////////////////////////////////////
    //                                                                  //
    //                            CONSTANTS                             //
    //                                                                  //
    //////////////////////////////////////////////////////////////////////

    /**
      * Version 1.0
      */
    public final static byte PDP_VERSION = 0x01;

    /**
      * Identifier for PDP ServiceRequest messages
      */
    protected final static byte PDP_SERVICE_REQUEST = 0x01;

    /**
      * Identifier for PDP ServiceReply messages
      */
    protected final static byte PDP_SERVICE_REPLY = 0x02;

    /**
      * Identifier for PDP ServiceDeregister messages
      */
    protected final static byte PDP_SERVICE_DEREGISTER = 0x03;

    /**
      * Minimun length in bytes for a PDP message
      */
    private final static byte MINIMUN_MESSAGE_LENGTH = 10;

    /**
      * Maximun length in bytes for a PDP message
      */
    private final static int MAXIMUN_MESSAGE_LENGTH = 0xFFFE;







    //////////////////////////////////////////////////////////////////////
    //                                                                  //
    //                    Message fields & flags                        //
    //                                                                  //
    //////////////////////////////////////////////////////////////////////

    /**
      * Function ID header field
      */
    private byte function_id;

    /**
      * Overflow flag. <code>true</code> whether the entry message does
      * not fit within one single datagram.
      */
    private boolean overflow;

    /**
      * "Any" message flag field. When <code>true</code> it specifies a
      * 1-request 1-reply type message
      */
    private boolean any;

    /**
      * XID header field. Acts as a message request identifier
      */
    private int xid;

    /**
      * ServiceType field from a ServiceRequest message
      */
    private String service_type;

    /**
     * ServiceEntryList/PRList field form a PDP message
     *
     * @uml.property name="entry"
     * @uml.associationEnd multiplicity="(0 -1)"
     */
    private ServiceEntry[] entry;






    /**
      * Constructor. This method creates a new PDP ServiceRequest message
      * with the information given on the parameters
      *
      * @param xid XID message field that identifies every
      *     ServiceRequest message
      *
      * @param service_type Type of service to discover on the network
      *
      * @param pr_list Vector of already known services of
      *     the type of service requested on the message
      *
      * @param any Any flag field that identifies tye type of the request
      *     sent to the network. When <code>true</code> this request
      *     is a 1-request 1-reply message type
      */
    protected PDPMessage ( int xid, String service_type, Vector pr_list,
                 boolean any ) throws PDPException
    {
        if ( service_type == null )
        {
            throw new PDPException (
                PDPException.PARSING_ERROR, "Searching for a null service" );
        }

        if ( pr_list == null ) pr_list = new Vector ( );
        int len = pr_list.size ( );

        this.any = any;

        if ( any ) len = 0;

        this.function_id = PDP_SERVICE_REQUEST;
        this.xid = xid;
        this.service_type = service_type;

        int length = 12 + service_type.length ( );

        for ( int i = 0; i < len; i ++ )
        {
            ServiceEntry e = (ServiceEntry)pr_list.elementAt (i);
            length += e.getLength ();
            if ( length >= MAXIMUN_MESSAGE_LENGTH )
            {
                len = i - 1;
                this.overflow = true;
                if ( len < 0 ) len = 0;
            }
        }

        newEntry ( pr_list, len );
    }

    /**
      * Constructor. Creates a new ServiceReply message with the information
      * given on the parameters. This method internaly uses
      * {@link PDPMessage#PDPMessage ( int, Vector, byte )} Constructor.
      *
      * @param xid XID message field that identifies every
      *     ServiceReply message
      *
      * @param entry Vector of services stored on the message
      */
    protected PDPMessage ( int xid, Vector entry ) throws PDPException
    {
        this ( xid, entry, PDP_SERVICE_REPLY );
    }


    /**
      * Constructor. Creates a new ServiceDeregister message with the
      * information given on the parameter.  This method internaly uses
      * {@link PDPMessage#PDPMessage ( Vector )} Constructor.
      *
      * @param entry Vector of services stored on the message
      */
    protected PDPMessage ( Vector entry ) throws PDPException
    {
        this ( 0x00, entry, PDP_SERVICE_DEREGISTER );
    }



    /**
      * Private Constructor. Creates a new
      * ServiceReply (see {@link PDPMessage#PDPMessage ( int, Vector )}) or
      * ServiceDeregister (see {@link PDPMessage#PDPMessage ( Vector )})
      * message to be sent throgh the network.
      *
      * @param xid XID message field that identifies every
      *     ServiceReply message. On a ServiceDeregister message
      *     this field is set to 0x00
      *
      * @param entry Vector of services stored on the message
      *
      * @param id FunctionID header field message
      */
    private PDPMessage ( int xid, Vector entry, byte id ) throws PDPException
    {
        int len = entry.size ( );
        int length = 10;

        this.function_id = id;
        this.any = false;
        this.xid = xid;

        for ( int i = 0; i < len; i ++ )
        {
            ServiceEntry e = (ServiceEntry)entry.elementAt(i);
            length += e.getLength ( );
            if ( length >= MAXIMUN_MESSAGE_LENGTH )
            {
                len = i - 1;
                this.overflow = true;
                if (len<0) len = 0;
            }
        }

        if ( len == 0 ) throw new PDPException ( PDPException.PARSING_ERROR );

        newEntry ( entry, len );
    }




    /**
      * Constructor. Creates a new PDPMessage object from a message received
      * from the network.
      *
      * @param data Byte array with the message received from the network
      *
      */
    protected PDPMessage ( byte [] data ) throws PDPException
    {
        //PDPUtil.showDatagramContent(data);
        // Does it have minimun message length?
        if ( data.length < MINIMUN_MESSAGE_LENGTH )

            throw new PDPException ( PDPException.PARSING_ERROR );


        // Does it have maximun message length?
        if ( data.length > MAXIMUN_MESSAGE_LENGTH )

            throw new PDPException ( PDPException.PARSING_ERROR );


        // Does it have a supported PDP version?
        if ( data[0] != PDP_VERSION )

            throw new PDPException ( PDPException.PARSING_ERROR );


        // Does it have a supported function ID?
        if ( data[1] < PDP_SERVICE_REQUEST )

            throw new PDPException ( PDPException.PARSING_ERROR );


        // Does it have a supported function ID?
        if ( data[1] > PDP_SERVICE_DEREGISTER )

            throw new PDPException ( PDPException.PARSING_ERROR );

        //
        // Well, message seems to be ok.. let's parse it
        //////////////////////////////////////////////////////////////////

        // Gets function ID
        this.function_id = data [1];

        // Gets Overflow flag
        this.overflow = ( data [2] & 0x80 ) == 0x80;

        // Gets Any_Message flag
        this.any = ( data [2] & 0x40 ) == 0x40;

        // Gets XID
        byte [] temp = new byte [2];
        System.arraycopy ( data, 6, temp, 0, 2 );
        this.xid = PDPUtil.toInt( temp );

        // Temporary length & data index
        int len;
        int index = 8;

        switch ( function_id )
        {
            case PDP_SERVICE_REQUEST:

                // Gets PR LIST LENTGTH
                temp = new byte [2];

                System.arraycopy ( data, index, temp, 0, 2 );
                len = PDPUtil.toInt ( temp );

                // Sets index on "PRList" starting point
                index += 2;

                // ServiceEntry length
                int _len2;

                // Temporary ServiceEntry store
                Vector _v = new Vector ( );

                // Bucle for getting service entries
                for ( int i = 0; i < len; i ++ )
                {
                    // Gets Service Entry Length
                    temp = new byte [2];
                    System.arraycopy ( data, index + 4, temp, 0, 2 );
                    _len2 = PDPUtil.toInt ( temp ) + 6;

                    // Checks for valid length
                    if ( data.length < index + _len2 )
                    {
                        throw new PDPException ( PDPException.PARSING_ERROR );
                    }

                    temp = new byte [_len2];

                    // Creates new Service Entry
                    System.arraycopy ( data, index, temp, 0, _len2);
                    try
                    {
                        ServiceEntry entry = new ServiceEntry ( temp );
                        _v.add ( entry );
                    }
                    catch ( PDPException e ) {}
                    index += _len2;
                }

                // Gets ServiceType LENGTH
                temp = new byte [2];
                System.arraycopy ( data, index, temp, 0, 2 );
                int lenaux = PDPUtil.toInt ( temp );
                index += 2;

                // Checks for valid length
                if ( data.length < index + lenaux )
                {
                    throw new PDPException ( PDPException.PARSING_ERROR );
                }

                // Gets ServiceType
                temp = new byte [lenaux];
                System.arraycopy ( data, index, temp, 0, lenaux );
                this.service_type = new String ( temp );

                newEntry (_v, len);
                break;


            case PDP_SERVICE_REPLY:
            case PDP_SERVICE_DEREGISTER:

                // Gets Number of Service Entry
                temp = new byte [2];
                System.arraycopy ( data, index, temp, 0, 2 );
                len = PDPUtil.toInt ( temp );

                // Sets index & service_entry_length
                index += 2;
                int len2;

                // Temporary ServiceEntry store
                Vector v = new Vector ( );

                // Bucle for getting service entries
                for ( int i = 0; i < len; i ++ )
                {
                    // Gets Service Entry Length
                    temp = new byte [2];
                    System.arraycopy ( data, index + 4, temp, 0, 2 );
                    len2 = PDPUtil.toInt ( temp ) + 6;

                    // Checks for valid length
                    if ( data.length < index + len2 )
                    {
                        throw new PDPException ( PDPException.PARSING_ERROR );
                    }

                    temp = new byte [len2];

                    // Creates new Service Entry
                    System.arraycopy ( data, index, temp, 0, len2);
                    try
                    {
                        ServiceEntry entry = new ServiceEntry ( temp );
                        v.add ( entry );
                    }
                    catch ( PDPException e ) {}
                    index += len2;
                }

                // Stores ServiceEntryList field
                newEntry ( v, v.size ( ) );
                v = null;
                break;
        }
    }

    /**
     * Returns whether the message is set as a 1-request 1-reply or not.
     * @return The "ANY" message flag field
     *
     * @uml.property name="any"
     */
    protected boolean isAny() {
        return any;
    }


    /**
      * Returns the ServiceEntry list within an array
      * @return The serviceentry message field from a ServiceReply or
      *     ServiceDeregister message
      */
    protected ServiceEntry[] getServiceEntry ( )
    {
        return entry;
    }

    /**
      * Returns the functionID header field of this PDP message.
      * @return FunctionID header field
      */
    protected byte getFunctionID ( )
    {
        return function_id;
    }


    /**
      * Returns the PRList field from a PDPServiceRequest message
      * @return PRList message field from a ServiceRequest message
      */
    protected ServiceEntry [] getPrList ( )
    {
        return entry;
    }

    /**
      * Returns the serviceType field from a PDPServiceRequest message
      * @return ServiceType message field from a ServiceRequest message
      */
    protected String getServiceType ( )
    {
        return service_type;
    }

    /**
      * Returns the xid header field from the PDP message.
      * @return XID header field
      */
    protected int getXID ( )
    {
        return xid;
    }

    /**
     * Sets the xid on this PDP message
     * @param xid The xid to set
     *
     * @uml.property name="xid"
     */
    protected void setXid(int xid) {
        this.xid = xid;
    }

    /**
      * Returns the flags header field form this PDP message
      * @return The flags header field
      */
    private byte getFlags ( )
    {
        byte data = 0x00;

        if ( overflow ) data |= 0x80;
        if (    any   ) data |= 0x40;

        return data;
    }


    /**
      * Translates this objecto into a byte array to send this PDP message
      * through the network
      */
    protected byte [] getBytes ( )
    {
        int length;
        byte [] data = null;
        byte [] entry = null;
        int j = this.entry.length;

        switch ( function_id )
        {
            case PDP_SERVICE_REQUEST:
                entry = getEntryBytes ( );
                int pr_len = entry.length;
                int sr_len = service_type.length ( );
                length = 12 + pr_len + sr_len;
                data = new byte [length];

                // Version
                data [0] = PDP_VERSION;

                // Function ID
                data [1] = function_id;

                // Flags
                data [2] = getFlags ( );

                // Message Length
                System.arraycopy ( PDPUtil.toByteArray( length, 3 ), 0,
                                   data, 3,
                                   3 );

                // XID
                System.arraycopy ( PDPUtil.toByteArray( xid, 2 ), 0,
                                   data, 6,
                                   2 );

                // PR List Length
                System.arraycopy ( PDPUtil.toByteArray ( j, 2), 0,
                                   data, 8,
                                   2);

                // PR List
                System.arraycopy ( entry, 0,
                                   data, 10,
                                   pr_len );

                // Service Type List
                System.arraycopy ( PDPUtil.toByteArray ( sr_len, 2), 0,
                                   data, 10 + pr_len,
                                   2 );

                // Service Type
                System.arraycopy ( service_type.getBytes ( ), 0,
                                   data, 12 + pr_len,
                                   sr_len );
                break;


            case PDP_SERVICE_REPLY:
            case PDP_SERVICE_DEREGISTER:
                length = 10;

                entry = getEntryBytes ( );
                int elength = entry.length;

                length += elength;

                data = new byte [length];

                // Version
                data [0] = PDP_VERSION;

                // Function ID
                data [1] = function_id;

                // Flags
                data [2] = getFlags ( );

                // Message Length
                System.arraycopy ( PDPUtil.toByteArray ( length, 3 ), 0,
                                   data, 3,
                                   3 );

                // XID
                System.arraycopy ( PDPUtil.toByteArray ( xid, 2 ), 0,
                                   data, 6,
                                   2 );

                // Number of URLEntries
                System.arraycopy ( PDPUtil.toByteArray (j, 2), 0,
                                   data, 8,
                                   2 );

                // URLEntries
                System.arraycopy ( entry, 0, data, 10, elength );
                entry = null;
                break;
        }

        return data;
    }

    /**
      * Fills <code>entry</code> field from the information given on
      * the parameters
      *
      * @param v Vector of {@link ServiceEntry} objects to add on the
      *     <code>entry</code> field
      *
      * @param i number of elements of the vector to add on the
      *     <code>entry</code> field
      */
    private void newEntry ( Vector v, int i )
    {
        entry = new ServiceEntry [i];

        for ( int j = 0; j < i; j++ )

            entry [j] = (ServiceEntry) v.elementAt ( j );
    }


    /**
      * Returns a byte array cotainning the information stored on
      * the <code>entry</code> field
      *
      * @return The byte array with the ServiceEntry message information
      *
      *
      */
    private byte [] getEntryBytes ( )
    {
        int j = this.entry.length;
        int elength = 0;

        for ( int i = 0; i < j; i++ )
        {
            elength += this.entry[i].getLength ( );
        }

        byte [] entry = new byte [elength];

        int iaux = 0;
        for ( int i = 0; i < j; i++ )
        {
            byte [] aux = this.entry[i].getBytes ( );
            int laux = aux.length;
            System.arraycopy ( aux, 0, entry, iaux, laux );
            iaux += laux;
        }

        return entry;
    }
}