//
//                     Pervasive Discovery Protocol
//                          Copyright (C) 2005
//
//    Jos Carlos Perea barreiro         Celeste Campo Vzquez
//      (jperea[at]it.uc3m.es)          (celeste[at]it.uc3m.es)
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
//
// Contact the authors at:
//
//    Jos Carlos Perea Barreiro (jperea [at]it.uc3m.es)
//    Celeste Campo Vazquez      (celeste[at]it.uc3m.es)
//
//        Universidad Carlos III de Madrid
//        Departamento de Ingeniera Telemtica
//        Av. Universidad, 30
//        Edif. Torres Quevedo.
//        (28911) Legans - Madrid - SPAIN
//
//
/////////////////////////////////////////////////////////////////////////////
package com.uc3m.it.pfc.pdp;

import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;

import com.uc3m.it.pfc.pdp.PDPException;
import com.uc3m.it.pfc.pdp.PDPMessage;

/**
 * @author jperea
 */

public class PDPNetworkManager
{
    public final static String  MULTICAST_ADDRESS = "239.255.255.253";
    public final static int     PORT              = 3000;  //427
    public final static int     WAITING_TIME      = 5000;  //5secs
    public final static int     MAX_PACKET_SIZE   = 65000;
    private InetAddress         group;

    /**
     * Checks network connectivity
     */
    private NetworkSniffer sniff = null;

    /**
     * Sends messages to the network
     */
    private NetworkWriter writer = null;

    /**
     * Reads messages from other devices
     */
    private NetworkReader reader = null;

    /**
     * Reference to PDP
     */
    private PDP pdp = null;

    /**
     * Flag where to set network connectivity
     */
    private boolean activenetwork = false;

    /**
     * Starts PDPNetworkManager object
     *
     * @param pdp
     *           reference to PDP
     * @param snifferLock
     *           monitor where to notify changes on the network connection
     */
    public PDPNetworkManager ( PDP pdp, Object snifferlock )
    {
        this.pdp  = pdp;
        this.sniff = new NetworkSniffer ( this, snifferlock );
    }

    private void kill ( )
    {
        writer.kill ( );
        reader.kill ( );
        writer           = null;
        reader           = null;
        activenetwork    = false;
        pdp.networkInformation(null);
    }

    protected void newData ( PDPMessage msg )
    {
        try
        {
            if ( activenetwork ) writer.newData ( msg );
        }
        catch ( Exception e )
        {
            // does nothing :)
        }
    }

    protected void startNM ( String ip ) throws Exception
    {
    	//System.out.println("[NM] [manag] [@" + System.currentTimeMillis ( ) + "] Starting network" );
        this.group = InetAddress.getByName ( MULTICAST_ADDRESS );
        writer = new NetworkWriter ( group ); //group, socket );
        reader = new NetworkReader ( pdp, ip, group ); //socket, ip, pdp );

        activenetwork = true;
        pdp.networkInformation(ip);
        //System.out.println("[NM] [manag] [@" + System.currentTimeMillis ( ) + "] Network up" );
    }

    protected void stopLoop ( )
    {
        try
        {
        	//System.out.println("[NM] [manag] [@" + System.currentTimeMillis ( ) + "] Stopping network" );
            kill ( );
            //System.out.println("[NM] [manag] [@" + System.currentTimeMillis ( ) + "] Network down" );
        }
        catch ( Exception e )
        {
        }
    }
    
    protected void stopSniffer ( )
    {
    	sniff.stopSniffer();
    }
}

class NetworkWriter extends Thread
{
    private DatagramPacket  packet;
    private Object          lock;
    private boolean         active;
    private byte []         data;
    private InetAddress     group;
    private MulticastSocket socket;

    NetworkWriter ( InetAddress group ) throws Exception //InetAddress group, MulticastSocket socket )
    {
    	super ( );
    	
        data = null;
        lock = new Object ( );
        active = true;
        packet = null;
        this.group = group;
        socket = new MulticastSocket ( PDPNetworkManager.PORT );
        this.setName("NETWORK_WRITER");
        
        this.start ( );
        //System.out.println("[NW] [write] [@" + System.currentTimeMillis ( ) + "] Starting network writer" );
    }

    void kill ( )
    {
        synchronized ( lock )
        {
        	//System.out.println("[NW] [write] [@" + System.currentTimeMillis ( ) + "] Closing network writer" );
        	socket.close ( );
            data   = null;
            group  = null;
            socket = null;
            active = false;
            lock.notify ( );
        }
    }

    protected void newData ( PDPMessage msg )
    {
        synchronized ( lock )
        {
            data = msg.getBytes ( );
            lock.notify ( );
        }
    }

    public void run ( )
    {
        byte [] senddata;
        int len;

        while ( active )
        {
            System.gc ( );

            synchronized ( lock )
            {
                if ( data == null )
                {
                    try
                    {
                        lock.wait ( );
                    }
                    catch ( Exception e )
                    {
                    }

                    // data may have changed during
                    // our waitting mode...
                    if ( data == null )
                    {
                        continue;
                    }
                }

                len = data.length;
                senddata = new byte [len];
                System.arraycopy ( data, 0, senddata, 0, len );
                data = null;
            }

            // so here there is new data to send
            packet = new DatagramPacket ( senddata,
                                          len,
                                          group,
                                          PDPNetworkManager.PORT );
            try
            {
                //PDPUtil.showDatagramContent(senddata);
                socket.send ( packet );
                //System.out.println("[NW] [write] [@" + System.currentTimeMillis ( ) + "] Packet sent" );
            }
            catch ( Exception e )
            {
                // does nothing
            }

            senddata = null;
        }

        data = null;
        lock = null;
        packet = null;

        System.gc ( );
        //System.out.println("[NW] [write] [@" + System.currentTimeMillis ( ) + "] Network writer is about to stop" );   
    }
}

class NetworkReader extends Thread
{
    private byte []         data;
    private DatagramPacket  packet;
    private boolean         active;
    private MulticastSocket socket;
    private String          localhostaddress;
    private InetAddress     group;

    /**
     *
     */
    private PDP pdp;

    NetworkReader ( PDP pdp, String localhostaddress, InetAddress group ) throws Exception //MulticastSocket socket, String localhostaddress, PDP pdp )
    {
    	super ( );
    	
        active = true;

        this.group  = group;
        this.socket = new MulticastSocket ( PDPNetworkManager.PORT );
        this.socket.joinGroup ( group );
        this.localhostaddress = localhostaddress;
        this.pdp              = pdp;

        this.setName("NETWORK_READER");
        
        this.start ( );
        //System.out.println("[NR] [read ] [@" + System.currentTimeMillis ( ) + "] Starting network reader" );
    }

    protected void kill ( )
    {
    	//System.out.println("[NR] [read ] [@" + System.currentTimeMillis ( ) + "] Closing network reader" );
    	try
    	{
    		socket.leaveGroup(group);
    	}
    	catch ( Exception e )
    	{}
    	
        active = false;
        socket.close();
        socket = null;
    }

    public void run ( )
    {
        while ( active )
        {
            System.gc ( );
           
            try
            {
                data   = new byte [PDPNetworkManager.MAX_PACKET_SIZE];
                packet = new DatagramPacket ( data, data.length );
            	//System.out.println("[NR] [read ] [@" + System.currentTimeMillis ( ) + "] Listening..." );
                socket.receive ( packet );
            }
            catch ( Exception e )
            {
                continue;
            }

            String address = packet.getAddress ( ).getHostAddress ( );

            synchronized ( localhostaddress )
            {
                if ( localhostaddress.equals ( address ) )
                {
                    continue;
                }
            }

            int len = packet.getLength ( );
            byte [] recv = new byte [len];
            System.arraycopy ( data, 0, recv, 0, len );

            //System.out.println("[NR] [read ] [@" + System.currentTimeMillis ( ) + "] New packet received" );
            
            PDPMessage msg;

            try
            {
                //PDPUtil.showDatagramContent(recv);
                msg = new PDPMessage ( recv );
            }
            catch ( PDPException e )
            {
                //System.out.println ( "Es aqui" );
                e.printStackTrace();
                continue;
            }

            recv = null;

            switch ( msg.getFunctionID ( ) )
            {
                case PDPMessage.PDP_SERVICE_REQUEST:
                	//System.out.println("[NR] [read ] [@" + System.currentTimeMillis ( ) + "] New service request received" );
                    pdp.add ( msg.getPrList ( ) );
                    pdp.newRequest ( msg );
                    break;

                case PDPMessage.PDP_SERVICE_REPLY:
                	//System.out.println("[NR] [read ] [@" + System.currentTimeMillis ( ) + "] New service reply received" );
                    pdp.add ( msg.getServiceEntry ( ) );
                    pdp.newReply ( msg );
                    break;

                case PDPMessage.PDP_SERVICE_DEREGISTER:
                	//System.out.println("[NR] [read ] [@" + System.currentTimeMillis ( ) + "] New service deregister received" );
                    pdp.remove ( msg.getServiceEntry ( ) );
                    break;

            // default does nothing
            // since there are no more
            // message types
            }
        }

        data   = null;
        packet = null;

        System.gc ( );
        //System.out.println("[NR] [read ] [@" + System.currentTimeMillis ( ) + "] Network reader is about to stop" );
    }
}

class NetworkSniffer extends Thread
{
    private boolean loop = false;
    private boolean activenetwork = false;

    /**
     *
     */
    private PDPNetworkManager padre;

    private Object lock;

    NetworkSniffer ( PDPNetworkManager padre, Object lock )
    {
    	super ( );
    	this.loop = true;
        this.padre = padre;
        this.lock = lock;
        this.setName("NETWORK_SNIFFER");
        this.start ( );        
        
        //System.out.println("[NS] [sniff] [@" + System.currentTimeMillis ( ) + "] Starting network sniffer" );
    }

    public void run ( )
    {
        System.gc ( );

        while ( loop )
        {
            try
            {
                synchronized ( this )
                {
                    red ( );
                    wait ( PDPNetworkManager.WAITING_TIME );
                }                
            }
            catch ( Exception e )
            {
            }
        }

        //System.out.println("[NS] [sniff] [@" + System.currentTimeMillis ( ) + "] Network sniffer is about to stop" );
        System.gc ( );
    }

    private void doError ( String ip )
    {
        if ( !activenetwork && ip != null )
        {
            padre.stopLoop ( );
            try
            {
                padre.startNM ( ip );
                activenetwork = true;

                synchronized ( lock )
                {
                    lock.notifyAll ( );
                    //System.out.println("[NS] [sniff] [@" + System.currentTimeMillis ( ) + "] Network is up!" );
                }
            }
            catch ( Exception e )
            {

            }
        }
        else if ( activenetwork && ip == null )
        {
            padre.stopLoop ( );
            activenetwork = false;

            synchronized ( lock )
            {
                lock.notifyAll ( );
                //System.out.println("[NS] [sniff] [@" + System.currentTimeMillis ( ) + "] Network is down!" );
            }
        }
    }

    private void red ( )
    {
        String ip = null;

        try
        {
            ip = InetAddress.getLocalHost ( ).getHostAddress ( );

            if ( ip.equals ( "0." ) || ip.startsWith ( "127." ) )
            {
                ip = null;
            }
        }
        catch ( Exception e )
        {
            ip = null;
        }

        doError ( ip );
    }

    public void stopSniffer ( )
    {
        synchronized ( this )
        {
        	//System.out.println("[NS] [sniff] [@" + System.currentTimeMillis ( ) + "] Closing network sniffer" );
            loop = false;
            notify ( );
        }
    }
}