//
//                     Pervasive Discovery Protocol
//                          Copyright (C) 2005
//
//    Jos Carlos Perea barreiro         Celeste Campo Vzquez
//      (jperea[at]it.uc3m.es)          (celeste[at]it.uc3m.es)
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
//
// Contact the authors at:
//
//    Jos Carlos Perea Barreiro (jperea [at]it.uc3m.es)
//    Celeste Campo Vazquez      (celeste[at]it.uc3m.es)
//
//        Universidad Carlos III de Madrid
//        Departamento de Ingeniera Telemtica
//        Av. Universidad, 30
//        Edif. Torres Quevedo.
//        (28911) Legans - Madrid - SPAIN
//
//
/////////////////////////////////////////////////////////////////////////////

/**
  * Provides the classes and interfaces relating to the reference
  * implementation for Pervasive Discovery Protocol (PDP).
  * <p>
  * This protocol was designed within the Departamento de Ingeniera
  * Telemtica at the Universidad Carlos III de Madrid by
  * Celeste Campo Vzquez (celeste[at]it.uc3m.es) and programmed by
  * Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
  */
package com.uc3m.it.pfc.pdp;

/**
  *
  * The PDPUtil class is an extra class that provides
  * two mechanisms to improve data management for this implementation.
  * <p>
  * It is completly unrelated with the protocol and has the solely
  * function of helping the programmer when managing messages and
  * data structures on this protocol.
  *
  *
  *
  * @author Celeste Campo Vzquez (celeste[at]it.uc3m.es)
  * @author Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
  * @version initial
  */
class PDPUtil
{
    /**
      * Provides a method to convert a int into a byte array, prepared
      * to be included into a network message.
      * <p>
      * Some examples are:
      * <pre>
      *    int       Array [0]   Array[1]    Array[2]    Array[3]
      *
      *    256 ---> [0000 0000] [0000 0000] [0000 0001] [0000 0000]
      *   1024 ---> [0000 0000] [0000 0000] [0000 0100] [0000 0010]
      *  65536 ---> [0000 0000] [0000 0001] [0000 0000] [0000 0000]
      * </pre>
      *
      * @param data Number to convert to byte array
      * @param len Length of the array where to enter the number
      *
      * @return Byte array with a binary representation of the
      *     number <code>data</code> or null when trying to
      *     retrieve an array with more than 4 elements
      */
    public static synchronized byte [] toByteArray ( int data, int len )
    {
        byte [] b;

        if ( len > 4 )
        {
            b = null;
        }
        else
        {
            b = new byte [len];

            for ( int i = 0; i < len; i++ )
            {
                b[i] = (byte) ( ( data >> ( 8 * ( len - (i+1) ) ) ) & 0xFF );
            }
        }

        return b;
    }


    /**
      * Converts a byte array with a binary representation of a number
      * into a int
      *
      * @param b Byte array to convert
      * @return Int with the numerical representation of this array
      */
    public static synchronized int toInt ( byte [] b )
    {
        int r = 0;
        int l = b.length;
        int i = 0;

        for ( i = 0; i < l; i ++ )
        {
            r = ( ( b[i] & 0xFF ) << ( 8 * ( l - (i+1) ) ) ) | r;
        }

        return r;
    }
    
    private final static String LINE =
        "  +-----------+-----------+-----------+-----------+";
    private final static String HEAD =
        "    0123 4567   0123 4567   0123 4567   0123 4567  ";


    private static synchronized byte [] toBinary ( byte b )
    {
        String binary = Integer.toBinaryString ( ((byte) b & 0xFF ) );
        int length = binary.length ();

        for ( int i = 0; i < 8 - length; i++ ) binary = "0" + binary;

        byte [] bytes = binary.getBytes ();
        for ( int i = 0; i < 8; i ++ )
            bytes[i] = (byte) ( bytes[i] - (byte)'0' );

        return bytes;
    }

    private static synchronized void showBlankByte ( byte b )
    {
        for ( ; b != 0; b--) System.out.print ( "**** **** | " );
    }
    private static synchronized void showBinaryByte ( byte [] b )
    {
        if ( b.length != 8 )
        {
            System.out.println ( "ERROR AL IMPRIMIR!!!" );
            return;
        }

        System.out.print ( b[0] + "" + b[1] + "" + b[2] + "" + b[3] + " " +
                           b[4] + "" + b[5] + "" + b[6] + "" + b[7] + " | " );
    }

    public static synchronized void showDatagramContent ( byte [] data )
    {
        int data_length = data.length;
        int cols = 4;
        int rows = data_length/4;

        System.out.println  ( );
        System.out.println  ( HEAD );
        for ( int i = 0; i < data_length; i ++ )
        {
            if ( i % cols == 0 )
            {
                System.out.println ( );
                System.out.println ( LINE );
                System.out.print ( "  | " );
            }

            showBinaryByte ( toBinary ( data[i] ) );

            if ( (i+1) == data_length )
                showBlankByte ( (byte)( ( cols - (i+1) % cols ) % cols ) );

            if ( (i+1)% cols == 0 || (i+1) == data_length )
            {
                i = i+1;

                // bytes a representar
                int max = max = i%cols;
                if ( i%cols == 0 ) max = cols;

                System.out.print ( " > " );
                for ( int b = i - max; b != i; b++ )
                {
                    char c = (char)data[b];
                    if ( data[b] < 33 || data[b] > 137 ) c = '.';
                    System.out.print ( c );
                }

                // quedan columnas vacias??
                for ( int b = 0; b < cols - max; b++ ) System.out.print ( " " );
                System.out.print ( " <" );

                System.out.print ( "--> " );
                for ( int b = i - max; b != i; b++ )
                {
                    String c = Integer.toHexString ( data[b] );
                    if ( c.length ( ) == 0 ) c = "00";
                    if ( c.length ( ) == 1 ) c = "0" + c;
                    System.out.print ( c + " " );
                }
                for ( int b = 0; b < cols - max; b++ ) System.out.print ( "   " );
                System.out.print ( "<" );

                i = i-1;
            }

        }
        System.out.println ( );
        System.out.println ( LINE );
    }
}
