//
//                     Pervasive Discovery Protocol
//                          Copyright (C) 2005
//
//    Jos Carlos Perea barreiro         Celeste Campo Vzquez
//      (jperea[at]it.uc3m.es)          (celeste[at]it.uc3m.es)
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
//
// Contact the authors at:
//
//    Jos Carlos Perea Barreiro (jperea [at]it.uc3m.es)
//    Celeste Campo Vazquez      (celeste[at]it.uc3m.es)
//
//        Universidad Carlos III de Madrid
//        Departamento de Ingeniera Telemtica
//        Av. Universidad, 30
//        Edif. Torres Quevedo.
//        (28911) Legans - Madrid - SPAIN
//
//
/////////////////////////////////////////////////////////////////////////////

/**
  * Provides the classes and interfaces relating to the reference
  * implementation for Pervasive Discovery Protocol (PDP).
  * <p>
  * This protocol was designed within the Departamento de Ingeniera
  * Telemtica at the Universidad Carlos III de Madrid by
  * Celeste Campo Vzquez (celeste[at]it.uc3m.es) and programmed by
  * Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
  */
package com.uc3m.it.pfc.pdp;

import java.util.Vector;

/**
 * The PDP_SA class implements the process that follows a Service Agent
 * on the PDP protocol when replying a ServiceRequest message.
 *
 * @author Celeste Campo Vzquez (celeste[at]it.uc3m.es)
 * @author Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
 * @version initial
 */

class PDP_SA extends Thread
{

    /**
     * Reference to PDP core, to access other agents and cache memory of
     * the device
     *
     * @uml.property name="padre"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private PDP padre = null;


    /**
      * Flag that indicates the type of request is being answered. When
      * <code>true</code> it is a 1-request 1-reply message.
      */
    private boolean any = false;

    /**
      * Type of service asked on the ServiceRequest
      */
    private String type;

    /**
      * Place to temporaly store services that are being sent as a reply
      */
    private Vector reply;

    /**
     * Place to store {@link ServiceEntry} objects (services) known from
     * the requester and/or sent by other devices
     *
     * @uml.property name="known"
     * @uml.associationEnd multiplicity="(0 -1)"
     */
    private ServiceEntry[] known;

    /**
     * Lock to manage synchronization with {@link PDP_SA#known}
     */
    private Object lock;

    /**
      * Constructor. Creates a new Service Agent (PDP_SA) thread to parse
      * and response a request received from network.
      *
      * @param padre Reference to PDP core to access cache memory and other
      *     agents of this device
      *
      * @param msg {@link PDPMessage} received with the ServiceRequest
      */
    protected PDP_SA ( PDP padre, PDPMessage msg )
    {
        super ( Integer.toString( msg.getXID() ) );

        this.padre = padre;

        this.type  = msg.getServiceType ( );
        this.reply = new Vector    (      );
        this.any   = msg.isAny     (      );
        this.known = msg.getPrList (      );
        this.lock  = new Object    (      );

        if ( this.known == null ) this.known = new ServiceEntry [0];

        //System.out.println("[SA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] SA active");
        this.start ( );
    }




    /**
      * Creates a new timing for this reply. This time is used to allow
      * other devices with more information to give the answer to the
      * requester.
      *
      * @return Milliseconds to wait for sending the reply to the
      *     requester
      */
    protected int getTiming ( )
    {
        //                                     DEFAULT_AVAIL_TIME
        //  max = CONFIG_WAIT_TX * ------------------------------------------
        //                           DEFAULT_AVAIL_TIME + ( TTL * Entries )
        //
        int max = PDP.CONFIG_WAIT_TRANSMIT *
                  PDP.DEFAULT_AVAILABILITY_TIME /
                ( PDP.DEFAULT_AVAILABILITY_TIME +
                ( padre.getTTL ( ) * reply.size ( ) ) );

        if ( max == 0 ) max  =  100;
        else            max *= 1000;

        //System.out.println("[SA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] Replying before " + max + "ms. " + reply.size() + "services" );
        return (int)(Math.random ( ) * max); // millisecs
    }





    /**
      * Core of this object. This method implements the work that a PDP
      * Service Agent follows to answer a request.
      */
    public void run ( )
    {
        // Gets data from Local Cache
        reply.addAll ( padre.getCache ( true ).getServices ( type ) );
        if ( !any )
        {
            // Gets data from remote cache
            reply.addAll ( padre.getCache ( false ). getServices ( type ) );
            delKnown ();
        }

        boolean reply_ua = true;
        boolean loop     = true;
        while ( loop )
        {
            loop = false;                   // Exit next time...

            if ( reply.size ( ) > 0 )
            {
                //System.out.println( "[SA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] Creating a reply" );
                reply_ua = true;            // Prepare new REPLY to UA
                int tmmm = getTiming ( );
                //System.out.println( "[SA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] timer generated... " + tmmm + " ms waitting" );

                synchronized ( lock )
                {
                    if ( known.length == 0 )
                    {
                        try
                        {
                            //System.out.println( "[SA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] listening" );
                            lock.wait ( tmmm );
                        }
                        catch ( Exception e )
                        {
                            reply_ua = false;
                            loop     = false;
                            known    = new ServiceEntry [0];
                        }

                        if ( known.length != 0 )    // if there is a reply
                        {
                            reply_ua = false;
                            if ( !any )
                            {
                                delKnown ();
                                loop = true;
                            }
                        }
                    }
                    else
                    {
                        reply_ua = false;

                        if ( !any )          // if it's a 1-N request...
                        {                    // ... remove known services
                            delKnown ( );
                            loop = true;
                        }
                        else                 // if it's a 1-1 request...
                        {                    // ... we dont reply
                            loop = false;
                        }
                    }
                }
            }
            else
            {
                //System.out.println( "[SA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] No information to send" );
                reply_ua = false;
                loop     = false;
            }
        }

        if ( reply_ua )
        {
            int xid = Integer.parseInt ( getName ( ) );

            try
            {
                PDPMessage msg = new PDPMessage ( xid, reply );
                //System.out.println( "[SA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] Sending reply to NetworkManager" );
                padre.toNetworkManager ( msg );
            }
            catch ( PDPException e ) {}
        }

        //System.out.println( "[SA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] SA dies" );
        padre.killSA ( getName ( ) );
    }


    /**
      * Removes from <code>Vector reply</code> the serices known by
      * the requester. Replies in PDP cannot be redundant, so any information
      * known by the requester, or sent by other devices must be removed form
      * the reply before sending the message
      */
    private void delKnown ( )
    {
        if ( known != null )
        {
            int j = reply.size ( ) - 1;
            for ( int i = j; i >= 0; i -- )
            {
                // Gets a PDPService from the "reply" Vector
                PDPService s1 =
                    ((ServiceEntry)reply.elementAt (i)).getPDPService();

                for ( int k = 0; k < known.length; k ++ )
                {
                    // Compare services
                    PDPService s2 = known[k].getPDPService();
                    if ( s1.equals ( s2 ) )
                    {
                        reply.remove(i);
                        k = known.length;
                    }
                }
            }

            known = new ServiceEntry [0];
        }
    }


    /**
      * Informs the thread that new external data has benn sent for this
      * request by other device, and stores this information on
      * field <code>known</code>.
      *
      * @param data Array of {@link ServiceEntry} objects sent by a
      *     external device as a reply for this request
      */
    protected void newData ( ServiceEntry [] data )
    {
        synchronized ( lock )
        {
            this.known = data;
            //System.out.println( "[SA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] New data ("+data.length+") arrives" );
            lock.notify ( );
        }
    }
}