//
//                     Pervasive Discovery Protocol
//                          Copyright (C) 2005
//
//    Jos Carlos Perea barreiro         Celeste Campo Vzquez
//      (jperea[at]it.uc3m.es)          (celeste[at]it.uc3m.es)
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
//
// Contact the authors at:
//
//    Jos Carlos Perea Barreiro (jperea [at]it.uc3m.es)
//    Celeste Campo Vazquez      (celeste[at]it.uc3m.es)
//
//        Universidad Carlos III de Madrid
//        Departamento de Ingeniera Telemtica
//        Av. Universidad, 30
//        Edif. Torres Quevedo.
//        (28911) Legans - Madrid - SPAIN
//
//
/////////////////////////////////////////////////////////////////////////////

/**
  * Provides the classes and interfaces relating to the reference
  * implementation for Pervasive Discovery Protocol (PDP).
  * <p>
  * This protocol was designed within the Departamento de Ingeniera
  * Telemtica at the Universidad Carlos III de Madrid by
  * Celeste Campo Vzquez (celeste[at]it.uc3m.es) and programmed by
  * Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
  */
package com.uc3m.it.pfc.pdp;

import java.util.Vector;

/**
 * The PDP_UA class implements the process that follows a PDP User Agent
 * when discovering services.<br><br>
 *
 *
 * @author Celeste Campo Vzquez (celeste[at]it.uc3m.es)
 * @author Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
 * @version 0.1
 */

class PDP_UA extends Thread
{

    /**
     * Reference to PDP core, to access other agents and cache memory of
     * the device
     *
     * @uml.property name="padre"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private PDP padre = null;

    /**
      * Type of service requested on this service discovery
      */
    private String type = null;

    /**
      * XID message identifier for a hypothetical ServiceRequest message
      */
    private int xid;

    /**
      * Flag that indicates the type of request asked by user; when
      * <code>true</code> it is a 1-request 1-reply process
      */
    private boolean any;

    /**
      * Temporal memory to store received replies from network
      */
    private Vector known;

    /**
     * Flag that indicates SD procces is enabled and running.
     */
    private boolean stop;

    /**
     * Lock for synchronization
     */
    private Object lock;


    /**
      * Constructor. Creates a new PDP User Agent thread to start the
      * service discovering process.
      *
      * @param padre {@link PDP} reference to PDP core.
      * @param xid int with the XID message field for the ServiceRequest message
      * @param any boolean that sets the requesting method to discover services.
      *     When <code>true</code> it specifies a 1-request 1-reply discovering
      *     process
      */
    protected PDP_UA ( PDP padre, String type, int xid, boolean any )
    {
        super ( Integer.toString ( xid ) );

        this.padre = padre;
        this.type  = type;
        this.xid   = xid;
        this.any   = any;
        this.known = new Vector ( );
        this.stop  = false;
        this.lock  = new Object ( );

        this.start ( );
        //System.out.println("[UA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] Starting UA for service " + type );
    }


    /**
      * Core of this object. This method implements the work that a PDP
      * User Agent follows when service discovering.
      */
    public void run ( )
    {
        known.addAll ( padre.getCache(true ).getServices(type) );
        known.addAll ( padre.getCache(false).getServices(type) );

        try
        {
            if ( ! ( any && ( known.size ( ) != 0 ) ) )
            {
                PDPMessage msg = new PDPMessage ( xid, type, known, any );

                //System.out.println("[UA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] Sending request to Network Manager" );
                padre.toNetworkManager ( msg );

                long timing = System.currentTimeMillis() + (PDP.CONFIG_WAIT_REPLY * 1000);

                while ( !stop && System.currentTimeMillis() <= timing )
                {
                    synchronized ( lock )
                    {
                        long l = timing - System.currentTimeMillis ( );
                        if ( l <= 0 ) break;

                        //System.out.println("[UA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] Listening for "+l+ "ms" );
                        lock.wait ( l );

                        if ( any ) timing = System.currentTimeMillis ( ) - 100;
                    }
                }
            }
            if ( !stop )
        	{
            	//System.out.println("[UA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] Reply sent to PDP" );
            	padre.reply ( known );
        	}
            //System.out.println("[UA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] User Agent is about to stop" );
        }
        catch ( PDPException e ) { }
        catch ( Exception    e ) { }
    }


    /**
      * Informs User Agent that new external data has been received, and
      * stores the new services on field <code>known</code>
      *
      * @param data Array of {@link ServiceEntry} objects with the information
      *     received from a ServiceReply message aswering the
      *     ServiceRequest message sent by User Agent
      */
    protected void newData ( ServiceEntry [] data )
    {
        synchronized ( lock )
        {
            for ( int i = 0; i < data.length; i ++ )
            {
                known.add ( data [ i ] );
            }

            //System.out.println("[UA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] New " + data.length + " elements arrived" );
            lock.notify();
        }
    }

    /**
     * {@link PDP} would use this method to stop SD proccess as a client's request
     */
    protected void stopUA ( )
    {
        synchronized ( lock )
        {
            stop = true;
            //System.out.println("[UA] [" + getName ( ) + "] [@" + System.currentTimeMillis ( ) + "] Stopping User Agent" );
            known.removeAllElements ( );
            lock.notify();
        }
    }
}