//
//                     Pervasive Discovery Protocol
//                          Copyright (C) 2005
//
//    Jos Carlos Perea barreiro         Celeste Campo Vzquez
//      (jperea[at]it.uc3m.es)          (celeste[at]it.uc3m.es)
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
//
// Contact the authors at:
//
//    Jos Carlos Perea Barreiro (jperea [at]it.uc3m.es)
//    Celeste Campo Vazquez      (celeste[at]it.uc3m.es)
//
//        Universidad Carlos III de Madrid
//        Departamento de Ingeniera Telemtica
//        Av. Universidad, 30
//        Edif. Torres Quevedo.
//        (28911) Legans - Madrid - SPAIN
//
//
/////////////////////////////////////////////////////////////////////////////

/**
  * Provides the classes and interfaces relating to the reference
  * implementation for Pervasive Discovery Protocol (PDP).
  * <p>
  * This protocol was designed within the Departamento de Ingeniera
  * Telemtica at the Universidad Carlos III de Madrid by
  * Celeste Campo Vzquez (celeste[at]it.uc3m.es) and programmed by
  * Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
  */
package com.uc3m.it.pfc.pdp;

/**
 * The ServiceEntry class implements the ServiceEntry structures used
 * on Pervasive Discovery Protocol and several method to manage and
 * parse the information stored on them.
 *
 * Version 0.1 Makes methods getLifetime, getPDPService and isNoCaching public
 *
 * @author Celeste Campo Vzquez (celeste[at]it.uc3m.es)
 * @author Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
 * @version 0.1
 */

public class ServiceEntry
{
    /**
      * Specifies the minimun ServiceEntry length (in bytes)
      */
    public static final byte MINIMUN_LENGTH = 6;

    /**
      * No_caching flag from a ServiceEntry. When <code>true</code> this
      * ServiceEntry wont be stored on any cache.
      */
    private boolean no_caching = false;

    /**
      * Expiration time (in seconds) for this Service
      */
    private long lifetime;

    /**
     * Service definition
     *
     * @uml.property name="service"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private PDPService service;



    /**
      * Constructor. Creates a new ServiceEntry from a Service definition
      * and a expiration time.
      *
      * @param service {@link PDPService} to store on this ServicEntry
      * @param lifetime Expiration time (in seconds) for this ServiceEntry
      */
    protected ServiceEntry ( PDPService service, long lifetime )
    {
        this.lifetime = lifetime;
        this.service  = service;
    }




    /**
      * Constructor. Creates a new ServiceEntry object from a byte array
      * received from the network within a PDP message
      *
      * @param data Byte array with a valid ServiceEntry structure. Otherwise
      *     the constructor will throw a {@link PDPException}
      */
    protected ServiceEntry ( byte [] data ) throws PDPException
    {
        // Is a valid data?
        if ( data == null )

            throw new PDPException ( PDPException.PARSING_ERROR );


        // Does it have the minimal length?
        if ( data.length < 6 )

            throw new PDPException ( PDPException.PARSING_ERROR );



        byte [] temp = new byte [2];

        // Gets Service Description Length
        System.arraycopy ( data, 4, temp, 0, 2 );
        int len = PDPUtil.toInt ( temp );

        // Does it have a valid length?
        if ( data.length != MINIMUN_LENGTH + len )

            throw new PDPException ( PDPException.PARSING_ERROR );



        // Sets Service Description Length
        byte [] description = new byte [len];

        // Gets No caching flag
        this.no_caching = ( ( data[0] & 0x80 ) == 0x80 );

        // Gets lifetime
        System.arraycopy ( data, 2, temp, 0, 2 );
        this.lifetime = (long) PDPUtil.toInt ( temp );

        // Gets Service Description
        System.arraycopy ( data, 6, description, 0, len );

        // Creates the new PDPService
        this.service = new ServiceURL ( new String ( description ) );
    }

    /**
      * Activates the NO_CACHING field to prohibit any device from adding
      * this ServiceEntry to its cache
      */
    protected void setNoCaching ( )
    {
        this.no_caching = true;
    }


    /**
      * Checks the value of the NO_CACHING flag
      *
      * @return A boolean value containning the value of the flag
      */
    public boolean isNoCaching ( )
    {
        return no_caching;
    }


    /**
      * Returns the service stored on this ServiceEntry
      *
      * @return The {@link PDPService} object
      */
    public PDPService getPDPService ( )
    {
        return service;
    }

    /**
     * Returns the expiration time for this Service
     *
     * @uml.property name="lifetime"
     */
    public int getLifetime() {
        return (int) lifetime;
    }

    /**
      * Translates this object to a byte array; a more friendly format to
      * send the ServiceEntry throgh the network
      *
      * @return A byte array containning the data and structure of
      *     this ServiceEntry
      */
    protected byte [] getBytes ( )
    {
        byte [] description = service.getDescription ( ).getBytes ( );
        int slen = description.length;

        byte [] data = new byte [slen + 6];

        // reserved... (see flags below)
        data[0] = 0x00;
        data[1] = 0x00;

        // flags
        if ( no_caching ) data [0] = (byte) ( data[0] | 0x80 );

        // Lifetime
        System.arraycopy ( PDPUtil.toByteArray((int)lifetime,2),0,data,2,2 );

        // Service Description Length
        System.arraycopy ( PDPUtil.toByteArray (slen, 2), 0, data, 4, 2 );

        // Service Description
        System.arraycopy ( description, 0, data, 6, slen );

        return data;
    }

    /**
      *  Returns the length (in bytes) of this ServiceEntry when it is
      * included on a PDPMessage (on a byte array format)
      */
    protected int getLength ( )
    {
        return MINIMUN_LENGTH + service.getDescription ( ).length ( );
    }
}