//
//                     Pervasive Discovery Protocol
//                          Copyright (C) 2005
//
//    Jos Carlos Perea barreiro         Celeste Campo Vzquez
//      (jperea[at]it.uc3m.es)          (celeste[at]it.uc3m.es)
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
//
// Contact the authors at:
//
//    Jos Carlos Perea Barreiro (jperea [at]it.uc3m.es)
//    Celeste Campo Vazquez      (celeste[at]it.uc3m.es)
//
//        Universidad Carlos III de Madrid
//        Departamento de Ingeniera Telemtica
//        Av. Universidad, 30
//        Edif. Torres Quevedo.
//        (28911) Legans - Madrid - SPAIN
//
//
/////////////////////////////////////////////////////////////////////////////

/**
  * Provides the classes and interfaces relating to the reference
  * implementation for Pervasive Discovery Protocol (PDP).
  * <p>
  * This protocol was designed within the Departamento de Ingeniera
  * Telemtica at the Universidad Carlos III de Madrid by
  * Celeste Campo Vzquez (celeste[at]it.uc3m.es) and programmed by
  * Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
  */
package com.uc3m.it.pfc.pdp;

/**
  *
  * The class ServiceURL implements the interface {@link PDPService} to
  * work as the Service definition model on this Pervasive Discovery
  * Protocol implementation. Further information about the structure
  * of a Service URL can be found on the PDP specification and on
  * <a href="http://www.faqs.org/rfcs/rfc2609.html" target=_blank>here</a>
  *
  *
  * @author Celeste Campo Vzquez (celeste[at]it.uc3m.es)
  * @author Jos Carlos Perea Barreiro (jperea[at]it.uc3m.es)
  * @version initial
  */
class ServiceURL implements PDPService
{

    /**
      * Service type identifier of this Service
      */
    private String service;


    /**
      * Address to reach this Service on this network
      */
    private String url;




    /**
      * Constructor. Creates a new ServiceURL by using a service type and
      * a service url string.
      *
      * @param service Service type identifier
      * @param url Address to reach the service
      *
      * @throws PDPException when:
      * <ul>
      * <li>Service type is <code>null</code> o has 0 length
      * <li>URL is <code>null</code> or has 0 length
      * <li>Service type is the string <code>"all"</code>
      * </ul>
      */
    protected ServiceURL ( String service, String url ) throws PDPException
    {
        if ( service == null )

            throw new PDPException (PDPException.PARSING_ERROR,"Null service");


        if ( url == null )

            throw new PDPException (PDPException.PARSING_ERROR,"Null url");


        int service_length = service.length ( );

        if ( service_length == 0 )

            throw new PDPException (PDPException.PARSING_ERROR,"Empty service");


        if ( url.length ( ) == 0 )

            throw new PDPException (PDPException.PARSING_ERROR,"Empty url");

        service = service.toLowerCase ( ).trim ( );
        url     =     url.toLowerCase ( ).trim ( );


        if ( service.startsWith ( "service:" ) )

            service = service.substring ( 8, service_length );

        if ( service.equals ( "all" ) || service.equals ( "" ) )

            throw new PDPException (
                PDPException.PARSING_ERROR, "Invalid service type" );

        this.service = service;
        this.url     = url;
    }


    /**
      * Constructor. Creates a new ServiceURL by using a service description
      *
      * @param description Univocal Service description
      *
      * @throws PDPException when:
      * <ul>
      * <li>Description is <code>null</code> o has 0 length
      * <li>Description does not start with <code>"service:"</code>
      * <li>Description does not contain either type or url information
      * </ul>
      */
    protected ServiceURL ( String description ) throws PDPException
    {
        // DESCRIPTION == "service:" + getService ( ) + "://" + getURL ( );

        if ( description == null )
        {
            throw new PDPException ( PDPException.PARSING_ERROR,
            "ServiceURL description is NULL!" );
        }

        description = description.toLowerCase ( );

        if ( !description.startsWith ( "service:" ) )
        {
            throw new PDPException ( PDPException.PARSING_ERROR,
            "ServiceURL " + description + " does not start with \"service:\"" );
        }

        int index;

        if ( ( index = description.indexOf ( "://" ) ) == -1 )
        {
            throw new PDPException ( PDPException.PARSING_ERROR,
            "ServiceURL " + description + " does not have a valid url" );
        }

        int len = description.length ( );

        if ( index <= 8 )
        {
            throw new PDPException ( PDPException.PARSING_ERROR,
            "ServiceURL " + description + " has a empty service_type" );
        }

        if ( ( index + 3 ) == len )
        {
            throw new PDPException ( PDPException.PARSING_ERROR,
            "ServiceURL " + description + " has a empty url" );
        }

        this.service = description.substring ( 8, index );
        this.url = description.substring ( index + 3, len );
    }

    /**
     * Returns the service type description of this Service
     * @see com.uc3m.it.pfc.pdp.PDPService#getService()
     *
     * @uml.property name="service"
     */
    public String getService() {
        return service;
    }

    /** Returns the address to reach this Service in the network
      * @see com.uc3m.it.pfc.pdp.PDPService#getURL()
      */
    public String getURL()
    {
        return url;
    }

    /** Returns <code>null</code> as ServiceURL does not support the
      * <code>attributes</code> field
      * @see com.uc3m.it.pfc.pdp.PDPService#getAttributes()
      */
    public String getAttributes()
    {
        return null;
    }

    /**
      * @see com.uc3m.it.pfc.pdp.PDPService#equals(String)
      */
    public boolean equals ( String service )
    {
        if ( service.equals ( "all" ) ) return true;

        //          "concrete".equals ( "concrete:abstract" ) -> FALSE
        // "concrete:whatever".equals ( "concrete:abstract" ) -> FALSE
        //
        //          "concrete".equals ( "concrete" ) -> TRUE
        // "concrete:abstract".equals ( "concrete" ) -> TRUE
        boolean resultado = false;

        if ( service.indexOf (':') != -1 )
        {
            resultado = service.equals ( getService ( ) );
        }
        else
        {
            int i = getService ( ).indexOf ( ':' );
            if ( i == -1 ) i = getService().length ();
            String concrete = getService ( ).substring ( 0, i );
            resultado = service.equals ( concrete );
        }

        return resultado;
    }

    /**
      * @see com.uc3m.it.pfc.pdp.PDPService#equals(PDPService)
      */
    public boolean equals(PDPService service)
    {
        return getDescription ( ).equals ( service.getDescription ( ) );
    }


    /**
      *
      * @see com.uc3m.it.pfc.pdp.PDPService#getDescription()
      */
    public String getDescription ( )
    {
        return "service:" + getService ( ) + "://" + getURL ( );
    }


    /**
      * @see com.uc3m.it.pfc.pdp.PDPService#getBytes()
      */
    public byte[] getBytes()
    {
        return getDescription ( ).getBytes ( );
    }
}